# ruff: noqa: F401

from . import api, models, types, warnings
from ._load_udf import load
from ._options import env as _env
from ._options import options
from ._public_api import ingest, ingest_nongeospatial
from ._run import run
from ._secrets import secrets
from ._submit import submit
from ._udf import udf
from ._utils import utils

# TODO: fused.{download, get_chunk_from_table, get_chunks_metadata} are deprecated here
from .core import cache, download, file_path, get_chunk_from_table, get_chunks_metadata
from .ipython.fused_magics import load_ipython_extension

# Version is set to 0.0.0 as the version string is automatically generated by
# https://github.com/mtkennerly/poetry-dynamic-versioning
__version__ = "1.20.0"

__all__ = [
    "api",
    "cache",
    "download",
    "file_path",
    "get_chunk_from_table",
    "get_chunks_metadata",
    "ingest",
    # "ingest_nongeospatial",
    "load",
    # "models",
    "run",
    # "secrets",
    "submit",
    # "types",
    "udf",
    "utils",
    # "warnings",
    # "options",
    # "_env",
    "load_ipython_extension",
]
