class FusedWarning(Warning):
    pass


class FusedDeprecationWarning(FusedWarning):
    pass


class FusedInProgressWarning(FusedWarning):
    pass


class FusedNonProductionWarning(FusedWarning):
    pass


class FusedOnPremWarning(FusedWarning):
    pass


class FusedIPythonWarning(FusedWarning):
    pass


class FusedUdfWarning(FusedWarning):
    pass


class FusedDefaultWarning(FusedWarning):
    pass


class FusedPythonVersionWarning(FusedWarning):
    pass


class FusedTypeWarning(FusedWarning):
    pass


class FusedPathWarning(FusedWarning):
    pass


class FusedReservedWarning(FusedWarning):
    pass


class FusedIgnoredWarning(FusedWarning):
    pass


class FusedImportWarning(FusedWarning):
    pass


class FusedRefreshingWarning(FusedWarning):
    pass
